import logging
from typing import List, Tuple

import numpy as np
from gcma.io.baseio import TValue
from gcma.io.param import TParams


class Prior:
    def __init__(
        self,
        mean: TValue = None,
        cov: np.ndarray = None,
        gamma: float = 1,
    ):
        """
        :param gamma: weigth of the regularisation term
        """
        self.mean = None if mean is None else np.array(mean)
        self.cov = None if cov is None else np.array(cov)
        self.gamma = gamma
        self.slice = np.s_[:]

    def get_distance(self, params: TParams) -> float:
        """
        Get the distance between current parameters and prior knowledge

        :param params: non normalised params of one sample
        """
        if self.mean is None or self.cov is None:
            return 0
        vals = np.array(params.vals)[self.slice]
        no = (vals - self.mean)[:, np.newaxis]
        return (no.T @ np.linalg.inv(self.cov) @ no)[0, 0] * self.gamma

    def update(self, results: np.ndarray):
        """
        Updates prior based on database parameters distribution

        :param results: parameters value for each member of the database, size (n_patients, n_params)
        """
        self.mean = results.mean(axis=0)
        self.cov = np.cov(results.T)

    @classmethod
    def from_rules(
        cls,
        rules: List[Tuple[str, float, float]],
        params: TParams,
        pop_size: int,
        **kwargs
    ):
        names = [x.name for x in params]
        idx, res = [], np.zeros((pop_size, len(rules)))
        for j, (p, mean, std) in enumerate(rules):
            idx.append(names.index(p))
            res[:, j] = np.random.normal(mean, std, pop_size)

        c = cls(**kwargs)
        c.update(res)
        c.slice = np.r_[idx]
        return c


log = logging.getLogger(__name__)
